#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Main entry point for gterminal based apps.

This is the entry point for running a cli frontend that speaks dbus and adds
and removes command objects dynamically. it uses the magic gterminal library
for history and dbus support in a mainloop. The needs of this client inspired
the creation of the gterminal library.
"""
# Copyright (C) 2009-2010  James Shubin, McGill University
# Written for McGill University by James Shubin <purpleidea@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from jhelp import logginghelp
from jhelp import misc			# for path_search
import os				# for path manipulations
import glib				# for xdg related functions
import collections			# for named tuple
import ConfigParser	# TODO: maybe replace with a better config parser ?
			# see: http://wiki.python.org/moin/ConfigParserShootout
import gterminal
import gloader
import gcmd
import prefix

# dbus
import dbus
import dbus.service	# for the dbus_service class
import dbus.mainloop.glib
dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)	# this has to be set
from jhelp import dbushelp

# xmlrpc
# TODO: import ???

# soap
# TODO: import ???

_ = lambda x: x			# add fake gettext function until i fix up i18n
__all__ = ('gportal', 'DBUS', 'XMLRPC', 'SOAP', 'SIGNAL', 'dbus_service')

# GLOBALS #####################################################################
# TODO: consider putting all this inside a globalize() function
# TODO: the function could return the found variable to the caller for logging.
# build named tuples for config settings because they are clean
config = ConfigParser.SafeConfigParser()
found = misc.path_search('gportalrc.conf',
	[
		os.path.join(glib.get_user_config_dir(), prefix.name()),
		os.path.join('/', 'etc', prefix.name()),
		prefix.prefix() and	# if installed, aka (we have a prefix!)
		os.path.join(prefix.prefix(), 'share', prefix.name()) or
		os.path.join('files'),	# if not installed, use local version
	]
)
assert found, _('Installation is missing a default configuration file.')
# TODO: log the "found" variable
config.read(found)

# dbus config settings
temp = collections.namedtuple('DBUS', ['load', 'bus', 'interface'])
DBUS = temp(
	config.getboolean('dbus', 'load'),
	config.get('dbus', 'bus'),
	config.get('dbus', 'interface')
)

# TODO: add XMLRPC config parameters
# xmlrpc config settings
temp = collections.namedtuple('XMLRPC', ['load'])
XMLRPC = temp(
	config.getboolean('xmlrpc', 'load')
)

# TODO: add SOAP config parameters
# soap config settings
temp = collections.namedtuple('SOAP', ['load'])
SOAP = temp(
	config.getboolean('soap', 'load')
)

# signal data
types = ['message', 'function', 'other']
temp = collections.namedtuple('SignalType', ['ARG0']+types)
SIGNAL = temp(
	'signalDispatcher',				# unique string constant
	*[str(x) for x in range(1, len(types)+1)]	# expand args with: star
)

# clean up the namespace
del config, found, temp, types, x







# TODO: gportal should support getting objects over dbus/dbustubes, xmlrpc, soap; using avahi where possible. local classes should be supported too.

# TODO: at some point later down the road, it might be nice to create a
# mechanism to allow mixedin classes to specify a command hierarchy; eg:
# command [sub-command [sub-sub-command] [args]
# a class might decide to provide a sub-command under an existing hierarchy.

# XXX: maybe we can add the manual dbus loading/unloading of commands in here ?
class gportal_gcmd_dbus:
	# XXX: eventually this dbus command should be a sub command of a
	# "loader" command, or mixed in with some generic loader command for
	# dbus, xmlrpc, and soap. (since the backend provider can be hidden
	# from the user, which doesn't *need* to know where the command comes
	# from)
	def do_dbus(self, argv=()):

		# XXX: see about how to get argparse to play nicely in here...
		if len(argv) > 0:
			if argv[0] == 'unload':
				# we could unload by:
				# 1) command name
				# 2) dbus bus_name:path
				# 3) some unique id
				return 'unload...'
			elif argv[0] == 'load':
				# we could load by:
				# 1) dbus bus_name:path
				# 2) ?
				# note: tab completion could be very useful for
				# typing these in... work on that ?
				return 'load...'

			return 'whoa'
		else:
			return 'whoa, this is gportal_gcmd_dbus!!'


	def do_optparse_test(self, argv=()):
		return 'testing optparse...'


class gportal:
	# XXX: split out the dbus specific functions into a separate class and
	# have this class inherit from it. we can build a separate class for
	# each backend that we add; one for dbus, soap, xmlrpc and so on...
	"""main frontend client class."""
	def __init__(self):

		# MISC ########################################################
		self.name = 'gportal'

		# LOGGING #####################################################
		self.__logging = logginghelp.logginghelp(
			name=self.name, wordymode=True, stderrlog=False,
			mylogpath=[], mykerning=1
		)
		self.log = self.__logging.get_log()	# main logger

		# OBJECTS #####################################################
		self.cmdobj = gloader.gloader([gcmd.base_gcmd])
		self.gterminal = gterminal.gterminal(
			self.name, cmdobj=self.cmdobj, ps1='%s> ' % self.name,
			logger=self.__logging.get_log('gterminal')
		)
		self.connected = {}

		# DBUS ########################################################
		self.iface = None


	# SIGNALS #############################################################
	def __signalNameOwnerChanged(self, *args, **kwargs):
		"""Signal handler for NameOwnerChanged."""
		# XXX: dbus signals, akin to other forms of programming
		# signals arrive at the signal handler with info params
		# specifying what they are and where they came from, to
		# allow the programmer to potentially write a "generic"
		# signal handler to cover many events. the main problem
		# with the dbus signal is that it doesn't include which
		# bus it came from; session, system, or otherwise. this
		# is important, since our method calls need to act on a
		# particular bus. as a workaround (until someone solves
		# this dilemna) assume that bus names are unique across
		# all buses, and therefore find the bus with this hack.
		bus = dbus.SessionBus()			# use a default
		for x in [dbus.SessionBus(), dbus.SystemBus()]:
			if kwargs['sender'] in dbushelp.list_buses(bus=x):
				bus = x
				break				# found

		# connect: (but avoid private buses)
		if len(args[2]) > 0 and len(args[1]) == 0 and not(args[0].startswith(':')):
			for (i, j, k) in self.dbus_find(bus=bus, namedBus=args[0]):
				self.dbus_connect(
					bus=bus,
					namedBus=i,
					objectPath=j,
					interface=k
				)

		# disconnect: (run when something leaves the bus)
		elif len(args[1]) > 0 and len(args[2]) == 0:
			# args[0] was disconnected
			self.dbus_disconnect(args[0])

		# TODO: are signal handlers supposed to return True or False?
		#return ???


	def signalDispatcher(self, *args, **kwargs):
		"""Generic signal handler to dispatch the various signal types
		to the correct handlers for them."""
		# XXX: it would be nice to be able to receive keywork arguments
		# when receiving dbus signals, however they aren't implemented.
		#t = kwargs.get('type', None)
		assert len(args) == 3, _('The signalDispatcher was expecting 3 args.')
		assert args[0] == SIGNAL.ARG0, _('The signalDispatcher was expecting a different arg0.')
		if args[1] == SIGNAL.message:
			self.__signalMessage(*args)
		elif args[1] == SIGNAL.function:
			pass
		else:
			msg = _('Signal: %s' % kwargs.get('member', '<unknown>'))
			self.log.info(msg)
			self.log.debug('Signal args: %s' % str(args))
			self.log.debug('Signal kwargs: %s' % str(kwargs))
			self.gterminal.msg(msg)

		# TODO: are signal handlers supposed to return True or False?
		#return ???


	def __signalMessage(self, *args, **kwargs):
		"""Message signal handler."""
		d = args[2]
		if isinstance(d.get('message', None), str):
			self.log.info(_('Message: %s') % d.get('message'))
			self.gterminal.msg(**d)
		else:
			self.log.warn(_('Signal message isn\'t str, it\'s: %s')
			% type(d.get('message', None)))


	# MAIN ################################################################
	def main(self):
		"""Main function call for the gportal class."""
		# run load(); either once the mainloop is up, or inline.
		#self.load()
		# XXX: move these imports to top once we know we're keeping them
		import gobject
		from jhelp import gobject_once
		gobject.once_add_seconds(0, self.load)

		self.gterminal.run()


	def load(self):
		"""Function to run once the mainloop is already running."""
		# from the global named tuple settings 
		if DBUS.load:
			if DBUS.bus == 'Session': bus = dbus.SessionBus()
			elif DBUS.bus == 'System': bus = dbus.SystemBus()
			else: bus = None
			self.load_dbus(bus=bus)		# run the load command!

		# TODO: add these...
		# if XMLRPC.load:
		#	...
		# if SOAP.load:
		#	...


	def load_dbus(self, bus=None):
		"""Get the dbus subsystem running."""
		# start by loading the dbus command
		if not(self.cmdobj._load(gportal_gcmd_dbus)):
			msg = _('Problem loading dbus.')
			self.log.error(msg)
			self.gterminal.msg(msg)
			return False

		# make dbus happy before connecting to the bus
		# TODO: this seems to need to be set before dbushelp is imported (WHY?)
		#dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)

		if bus is None:
			bus = dbus.SessionBus()
			msg = _('DBus defaulting to Session bus.')
			self.log.info(msg)
			self.gterminal.msg(msg)

		# attach to main dbus event signals
		remote_object = bus.get_object(dbus.BUS_DAEMON_NAME, dbus.BUS_DAEMON_PATH)
		self.iface = dbus.Interface(remote_object, dbus.BUS_DAEMON_IFACE)
		# NameOwnerChanged (STRING name, STRING old_owner, STRING new_owner)
		# This signal indicates that the owner of a name has changed.
		# It's also the signal to use to detect the appearance of new names on the bus.
		# http://dbus.freedesktop.org/doc/dbus-specification.html#message-bus-messages
		self.iface.connect_to_signal('NameOwnerChanged', self.__signalNameOwnerChanged, **dbushelp.DBUS_SIGNAL_KEYWORDS)
		# TODO: do i need these two signals ?
		#self.iface.connect_to_signal('NameLost', self.signalhandler, **dbushelp.DBUS_SIGNAL_KEYWORDS)
		#self.iface.connect_to_signal('NameAcquired', self.signalhandler, **dbushelp.DBUS_SIGNAL_KEYWORDS)

		# find the magic interfaces
		for (i, j, k) in self.dbus_find(bus=bus):
			self.dbus_connect(
				bus=bus,
				namedBus=i,
				objectPath=j,
				interface=k
			)
		return True


	def dbus_find(self, bus=None, namedBus=None):
		"""Return an array of (namedBus, objectPath) tuples which
		contain the magic interface. Limit search to a particular
		namedBus if argument is specified. This obeys allow/deny."""
		result = []

		if bus is None:
			bus = dbus.SessionBus()

		if namedBus is None:
			names = dbushelp.list_buses(bus=bus)
		else:
			names = [namedBus]

		# TODO: we could maybe slim this big nested mess down with a
		# more direct dbus call, however for now, this is most logical.
		for i in names:
			#if i not in ???: continue	# TODO: add allow/deny list
			for j in dbushelp.list_objects(namedBus=i):
				for k in dbushelp.list_interfaces(namedBus=i, objectPath=j):
					# found magic interface
					if k == DBUS.interface:
						result.append((i, j, k))
		return result


	def dbus_connect(self, bus, namedBus, objectPath, interface=None):
		"""Actually connect in a chunk of dbus exported functions. This
		will grab at anything found on the interface. Interface default
		is the magic dbus ginterface name if None or not specified."""
		if interface is None: interface = DBUS.interface
		# use real str dict keys
		i = str(namedBus)
		j = str(objectPath)
		k = interface		# renamed to k for brevity

		# functions
		function_names = dbushelp.list_methods(
			namedBus=i,
			objectPath=j,
			interface=k
		)

		# get the actual proxy functions from their names
		# NOTE: we *could* do these three lines with a single call to
		# dbushelp.list_methods(..., returnObjects=True) however this
		# extra processing is repeated here so that we get one common
		# call to list the functions if between calls there's a change.
		remote_object = bus.get_object(i, j)
		iface = dbus.Interface(remote_object, k)
		functions = [iface.get_dbus_method(x) for x in function_names]

		# NOTE: functions that don't start with do_() are still
		# exported to this class and can be called, even though
		# they aren't publicly seen.
		new_class = gloader.build_class(functions)
		result = self.cmdobj._load(new_class)
		if result:
			# signals
			# add in the signals once we know the functions worked!
			signal_names = dbushelp.list_signals(
				namedBus=i,
				objectPath=j,
				interface=k
			)
			# XXX: maybe this could link up to sig_XYZ() functions?
			for l in signal_names:
				remote_object = bus.get_object(i, j)
				iface = dbus.Interface(remote_object, k)
				# XXX: generic signal handler for now... ideas?
				# all signals get sent to a generic dispatcher-
				# handler, which then routes them to a specific
				# signal handler; one designated for each type.
				# NOTE: i could dispatch with arg0='somestring'
				# but it's easier to dispatch it with one func.
				# REFERENCE: http://lists.freedesktop.org/archives/dbus/2006-December/006725.html
				iface.connect_to_signal(
					l,
					self.signalDispatcher,
					# so isinstance(dbus.String, str) works
					utf8_strings=True,
					**dbushelp.DBUS_SIGNAL_KEYWORDS
				)

			# store a ref to the class
			# dict:bus_name, dict:path
			self.connected.setdefault(i, {})
			self.connected[i].setdefault(j, {})
			# we don't need to store the list of function names
			# since they can be retrieved using: dir(new_class)
			self.connected[i][j] = new_class
			msg = _('Connected: %s:%s') % (i, j)
			self.log.info(msg)
			self.gterminal.msg(msg)
			# TODO: at some point we might want to save or display
			# this list of names so that an eager user can know
			# that a particular function appeared or dissapeared.
			return function_names	# in case someone wants them...
		else:
			self.log.warn(_('Problem connecting: %s:%s.') % (i, j))
			return []


	def dbus_disconnect(self, namedBus, objectPath=None):
		"""Actually disconnect a chunk of dbus functions. This takes a
		namedBus which may or may not still exist on the bus."""
		# TODO: do we need to disconnect signals too!?
		# check to see which paths were loaded; unload each one
		result = True
		for j,value in self.connected.get(namedBus, {}).items():
			# if objectPath is None, then disconnect from all the
			# paths on the namedBus, otherwise only disconnect from
			# the specified path.
			if objectPath is not None and objectPath != j: continue
			if self.cmdobj._unload(value):
				msg = _('Disconnected: %s:%s') % (namedBus, j)
				self.log.info(msg)
				self.gterminal.msg(msg)
			else:
				# didn't unload; likely wasn't loaded to begin.
				msg = _('Problem disconnecting: %s:%s') % (namedBus, j)
				self.log.warn(msg)
				# unnecessary to send a gterminal msg for this.
				result = False
		return result


# XXX: write a blog post about this, and how it is easy for a programmer to
# create a dbus service with this base class and how it inherits magic methods.
class dbus_service(dbus.service.Object):
	"""Utility class to be inherited by any children classes that
	wish to be dbus services. It's not required, but it does
	provide some useful functions."""
	def __init__(self, bus=dbus.SessionBus(),
		# TODO: is there a better default busname ?
		busname='ca.mcgill.cs.dazzle.gportal.service', path='/'):
		"""Constructor with some relatively sane defaults."""
		# the use of the getattr lets a variable specified in the child
		# class definition get used overriding what __init__() supplied
		self.__dbus_bus = getattr(self, 'bus', bus)
		self.__bus_name = dbus.service.BusName(
			getattr(self, 'busname', busname),
			bus=self.__dbus_bus
		)
		dbus.service.Object.__init__(
			self,
			self.__bus_name,
			getattr(self, 'path', path)
		)


	# specify the constant for the magic interface...
	@dbus.service.signal(DBUS.interface, signature='ssa{sv}')
	def __message(self, arg0, signal, data):
		"""Signal function to be used to send messages to gportal."""
		pass


	# using one underscore so this doesn't conflict with the child
	def _message(self, message, lifetime=None, name=None):
		"""Causes a message to be sent to the gterminal instance."""
		data = {'message': message}
		if lifetime is not None: data.setdefault('lifetime', lifetime)
		if name is not None: data.setdefault('name', name)
		return self.__message(SIGNAL.ARG0, SIGNAL.message, data)


if __name__ == '__main__':
	obj = gportal()
	obj.main()

